import sys

import pendulum
from sortedcontainers import SortedDict
from tbm_utils import (
	filter_filepaths_by_dates,
	get_filepaths,
)

from .config import (
	read_config_file,
	write_config_file,
)
from .core import (
	create_dir_info_dict,
	create_file_info_dict,
	read_torrent_file,
	write_torrent_file,
)
from .output import (
	generate_abbreviations_outputs,
	generate_magnet_link,
	generate_magnet_outputs,
	generate_summary_outputs,
	render,
)
from .utils import (
	calculate_data_size,
	calculate_piece_size,
	generate_unique_string,
)


def do_abbrs(args):
	conf = read_config_file()

	if args._subcommand == 'add':
		conf['trackers'][args.abbreviation] = args.tracker
	elif args._subcommand in ['rem', 'remove']:
		for abbreviation in args.abbreviations:
			try:
				del conf['trackers'][abbreviation]
			except KeyError:
				pass

	write_config_file(conf)
	outputs = generate_abbreviations_outputs(conf)
	render(outputs)


def do_create(args):
	torrent_info = SortedDict()

	filepaths = get_filepaths(
		args.input,
		max_depth=args.max_depth,
		exclude_paths=args.exclude_paths,
		exclude_regexes=args.exclude_regexes,
		exclude_globs=args.exclude_globs
	)

	creation_dates = [
		args[option]
		for option in [
			'created_in',
			'created_on',
			'created_before',
			'created_after',
		]
		if option in args
	]

	modification_dates = [
		args[option]
		for option in [
			'modified_in',
			'modified_on',
			'modified_before',
			'modified_after',
		]
		if option in args
	]

	filepaths = filter_filepaths_by_dates(
		filepaths,
		creation_dates=creation_dates,
		modification_dates=modification_dates,
	)

	filepaths = list(filepaths)

	if not filepaths:
		sys.exit("\nNo files matching criteria found.")

	data_size = calculate_data_size(filepaths)
	piece_size = calculate_piece_size(data_size, threshold=args.piece_threshold)

	if not args.trackers:
		private = False
	else:
		private = args.private

	if args.input.is_dir():
		info_dict = create_dir_info_dict(
			args.input,
			filepaths,
			data_size,
			piece_size,
			private,
			args.source,
			args.md5,
			show_progress=args.show_progress,
		)
	elif args.input.is_file():
		info_dict = create_file_info_dict(
			filepaths,
			data_size,
			piece_size,
			private,
			args.source,
			args.md5,
			show_progress=args.show_progress,
		)

	torrent_info['info'] = info_dict

	if args.trackers:
		torrent_info['announce'] = args.trackers[0][0]

		if len(args.trackers) > 1 or len(args.trackers[0]) > 1:
			torrent_info['announce-list'] = args.trackers

	if args.created_by is not None:
		torrent_info['created by'] = args.created_by

	if args.comment is not None:
		torrent_info['comment'] = args.comment

	torrent_info['creation date'] = pendulum.now('utc').int_timestamp

	torrent_info['encoding'] = 'UTF-8'

	write_torrent_file(args.output, torrent_info)

	outputs = generate_summary_outputs(torrent_info, show_files=args.show_files)
	render(outputs)


def do_info(args):
	torrent_info = read_torrent_file(args.torrent)

	outputs = generate_summary_outputs(torrent_info, show_files=args.show_files)
	render(outputs)


def do_magnet(args):
	torrent_info = read_torrent_file(args.torrent)
	magnet_link = generate_magnet_link(torrent_info)

	outputs = generate_magnet_outputs(magnet_link)
	render(outputs)


def do_xseed(args):
	torrent_info = read_torrent_file(args.torrent)

	if (
		not isinstance(torrent_info, dict)
		or 'info' not in torrent_info
	):
		raise ValueError(
			f"{args.torrent} is not a valid torrent file."
		)

	for k in ['announce-list', 'comment']:
		torrent_info.pop(k, None)

	torrent_info['info']['salt'] = generate_unique_string()

	if not args.trackers:
		torrent_info['info']['private'] = 0
	elif args.private is not None:
		torrent_info['info']['private'] = 1 if args.private else 0

	torrent_info['info'].pop('source', None)
	if args.source is not None:
		torrent_info['info']['source'] = args.source

	if args.trackers:
		torrent_info['announce'] = args.trackers[0][0]

		if len(args.trackers) > 1 or len(args.trackers[0]) > 1:
			torrent_info['announce-list'] = args.trackers
	else:
		torrent_info.pop('announce', None)
		torrent_info.pop('announce-list', None)

	if args.created_by is not None:
		torrent_info['created by'] = args.created_by

	if args.comment is not None:
		torrent_info['comment'] = args.comment

	torrent_info['creation date'] = pendulum.now('utc').int_timestamp

	torrent_info['encoding'] = 'UTF-8'

	write_torrent_file(args.output, torrent_info)

	outputs = generate_summary_outputs(torrent_info)
	render(outputs)



import math
import random
from pathlib import PurePath

import pendulum
from rich.bar import Bar
from rich.console import Console
from rich.progress import (
	BarColumn,
	Progress,
	ProgressColumn,
)
from rich.table import Table
from rich.text import Text
from tbm_utils import (
	cast_to_list,
	humanize_duration,
	humanize_filesize,
)

from .config import CONFIG_PATH
from .constants import DEFAULT_ABBRS
from .utils import (
	calculate_torrent_size,
	hash_info_dict,
)


CONSOLE = Console()


class BarColumn(BarColumn):
	def render(self, task):
		return Bar(
			total=task.total,
			completed=task.completed,
			width=self.bar_width,
			style='dim white',
		)


class ElapsedColumn(ProgressColumn):
	def render(self, task):
		elapsed = task.elapsed
		if elapsed is None:
			return Text("00:00", style="bold cyan")
		return Text(f" {humanize_duration(elapsed)}", style="bold cyan")


class ETAColumn(ProgressColumn):
	# Only refresh twice a second to prevent jitter
	max_refresh = 0.5

	def render(self, task):
		remaining = task.time_remaining
		if remaining is None:
			return Text("?", style="bold cyan")
		return Text(f"({humanize_duration(remaining)})", style="bold cyan")


class HashingSpeedColumn(ProgressColumn):
	max_refresh = 0.5

	def render(self, task):
		speed = task.speed
		if speed is None:
			return Text("?", style="bold red")
		hashing_speed = humanize_filesize(speed)
		return Text(f"{hashing_speed}/s", style="bold red")


PROGRESS = Progress(
	' ',
	'[bold magenta]{task.percentage:>3.0f}%',
	BarColumn(bar_width=60),
	HashingSpeedColumn(),
	'•',
	ElapsedColumn(),
	ETAColumn(),
	refresh_per_second=5,
)


def generate_abbreviations_outputs(conf):
	outputs = ['\n']

	config_table = Table(
		box=None,
		show_footer=False,
		show_edge=False,
		header_style="bold yellow underline",
	)
	config_table.add_column(
		'File',
		style='cyan',
	)
	config_table.add_row(None)
	config_table.add_row(str(CONFIG_PATH))

	outputs.append(config_table)

	auto_abbrs_table = Table(
		box=None,
		show_footer=False,
		show_edge=False,
		header_style="bold yellow underline",
	)

	auto_abbrs_table.add_column(
		'Auto',
		style='yellow',
	)
	auto_abbrs_table.add_column(style='cyan')

	auto_abbrs_table.add_row(None)
	auto_abbrs_table.add_row(
		'open',
		'All default trackers in a random tiered order.',
	)
	auto_abbrs_table.add_row(
		'random',
		'A single random default tracker.',
	)

	outputs.extend(['\n', auto_abbrs_table])

	default_abbrs_table = Table(
		box=None,
		show_footer=False,
		show_edge=False,
		header_style="bold yellow underline",
	)

	default_abbrs_table.add_column(
		'Default',
		style='yellow',
	)
	default_abbrs_table.add_column(style='cyan')

	default_abbrs_table.add_row(None)
	for abbr, tracker in DEFAULT_ABBRS.items():
		default_abbrs_table.add_row(abbr, tracker)

	outputs.extend(['\n', default_abbrs_table])

	user_abbrs = conf.get('trackers')
	if user_abbrs:
		user_abbrs_table = Table(
			box=None,
			show_footer=False,
			show_edge=False,
			header_style="bold yellow underline",
		)

		user_abbrs_table.add_column(
			'User',
			style='yellow',
		)
		user_abbrs_table.add_column(style='cyan')

		user_abbrs_table.add_row(None)
		for abbr, tracker in user_abbrs.items():
			user_abbrs_table.add_row(abbr, tracker)

		outputs.extend(['\n', user_abbrs_table])

	return outputs


def generate_magnet_link(torrent_info):
	torrent_name = torrent_info['info']['name']
	info_hash = hash_info_dict(torrent_info['info'])
	data_size = calculate_torrent_size(torrent_info)

	magnet_link = f'magnet:?dn={torrent_name}&xt=urn:btih:{info_hash}&xl={data_size}'

	if 'announce-list' in torrent_info:
		for tier in torrent_info['announce-list']:
			magnet_link += f'&tr={random.choice(tier)}'
	elif 'announce' in torrent_info:
		magnet_link += f'&tr={torrent_info["announce"]}'

	return magnet_link


def generate_magnet_outputs(magnet_link):
	outputs = ['\n']

	magnet_table = Table(
		box=None,
		show_footer=False,
		show_edge=False,
		header_style="bold yellow underline",
	)

	magnet_table.add_column(
		'Magnet',
		style='cyan',
	)

	magnet_table.add_row(None)
	magnet_table.add_row(magnet_link)

	outputs.append(magnet_table)

	return outputs


def generate_summary_outputs(torrent_info, show_files=False):
	outputs = ['\n']

	torrent_name = torrent_info['info']['name']
	info_hash = hash_info_dict(torrent_info['info'])
	private = 'Yes' if torrent_info['info'].get('private') == 1 else 'No'

	announce_list = None
	if 'announce-list' in torrent_info:
		announce_list = torrent_info['announce-list']
	elif 'announce' in torrent_info:
		announce_list = [[torrent_info['announce']]]

	data_size = calculate_torrent_size(torrent_info)
	piece_size = torrent_info['info']['piece length']
	piece_count = math.ceil(data_size / piece_size)

	tz = pendulum.tz.local_timezone()
	creation_date = pendulum.from_timestamp(
		torrent_info['creation date'],
		tz
	).format('YYYY-MM-DD HH:mm:ss Z')
	created_by = torrent_info.get('created by', '')
	comment = torrent_info.get('comment', '')
	source = torrent_info.get('source', '')

	magnet_link = generate_magnet_link(torrent_info)

	summary_table = Table(
		box=None,
		show_footer=False,
		show_edge=False,
		header_style="bold yellow underline",
	)

	summary_table.add_column(
		'Summary',
		style='yellow',
		no_wrap=True,
	)
	summary_table.add_column(style='cyan')

	summary_table.add_row(None)
	summary_table.add_row('Info Hash:', info_hash)
	summary_table.add_row('Torrent Name:', torrent_name)
	summary_table.add_row('Data Size:', humanize_filesize(data_size, precision=2))
	summary_table.add_row('Piece Size:', humanize_filesize(piece_size))
	summary_table.add_row('Piece Count:', str(piece_count))
	summary_table.add_row('Private:', str(private))
	summary_table.add_row('Creation Date:', str(creation_date))
	summary_table.add_row('Created By:', created_by)
	summary_table.add_row('Comment:', comment)
	summary_table.add_row('Source:', source)
	summary_table.add_row('Magnet:', magnet_link)

	outputs.append(summary_table)

	tracker_table = Table(
		box=None,
		show_footer=False,
		show_edge=False,
		header_style="bold yellow underline",
	)
	tracker_table.add_column(
		'Trackers',
		style='yellow',
		no_wrap=True,
	)
	tracker_table.add_column(style='cyan')
	tracker_table.add_row(None)

	if not announce_list:
		tracker_table.add_row('[cyan]None')
	else:
		pad = len(str(len(announce_list)))
		for i, tier in enumerate(announce_list):
			tracker_table.add_row(f"Tier {i:>{pad}}:", '\n'.join(tracker for tracker in tier))

	outputs.extend(['\n', tracker_table])

	if show_files:
		file_infos = []
		if 'files' in torrent_info['info']:
			for f in torrent_info['info']['files']:
				file_infos.append(
					(
						humanize_filesize(f['length'], precision=2),
						PurePath(*f['path']),
					)
				)
		else:
			file_infos.append(
				(
					humanize_filesize(
						torrent_info['info']['length'], precision=2
					),
					PurePath(torrent_info['info']['name']),
				)
			)

		num_pad = len(
			max(
				(
					size.split()[0]
					for size, _ in file_infos
				),
				key=len,
			)
		)
		unit_pad = len(
			max(
				(
					size.split()[1]
					for size, _ in file_infos
				),
				key=len,
			)
		)

		files_table = Table(
			box=None,
			show_footer=False,
			show_edge=False,
			header_style="bold yellow underline",
		)
		files_table.add_column(
			'Files',
			style='yellow',
			no_wrap=True,
		)
		files_table.add_column(style='cyan')
		files_table.add_row(None)

		for size, path in file_infos:
			num, unit = size.split()
			files_table.add_row(f"{num:>{num_pad}} {unit:>{unit_pad}}", str(path))

		outputs.extend(['\n', files_table])

	return outputs


@cast_to_list
def render(outputs, **kwargs):
	for output in outputs:
		CONSOLE.print(output, **kwargs)



from collections import ChainMap
from pathlib import Path

import appdirs
from sortedcontainers import SortedDict
from tomlkit.toml_document import TOMLDocument
from tomlkit.toml_file import TOMLFile

from .__about__ import __author__, __title__
from .constants import DEFAULT_ABBRS

CONFIG_PATH = Path(appdirs.user_config_dir(__title__, __author__), 'thorod.toml')


def read_config_file():
	config_file = TOMLFile(CONFIG_PATH)
	try:
		config = config_file.read()
	except FileNotFoundError:
		config = TOMLDocument()

	if 'trackers' not in config:
		config['trackers'] = SortedDict()

	write_config_file(config)

	return config


def write_config_file(config):
	CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
	CONFIG_PATH.touch()

	config_file = TOMLFile(CONFIG_PATH)
	config_file.write(config)


ABBRS = ChainMap(DEFAULT_ABBRS, read_config_file()['trackers'])



import argparse
import math
from pathlib import Path

import colorama
from tbm_utils import (
	Namespace,
	SubcommandHelpFormatter,
	UsageHelpFormatter,
	create_parser_filter_dates,
	create_parser_local,
	create_parser_meta,
	custom_path,
	datetime_string_to_time_period,
	get_defaults,
	merge_defaults,
	parse_args
)

from . import __title__, __version__
from .commands import (
	do_abbrs,
	do_create,
	do_info,
	do_magnet,
	do_xseed,
)
from .config import ABBRS, read_config_file
from .constants import (
	DEFAULT_ABBRS,
	DEFAULT_TRACKERS,
	PIECE_SIZE_STRINGS,
)

COMMAND_KEYS = {
	'abbrs',
	'create',
	'info',
	'magnet',
	'xseed',
}

colorama.init()


#########
# Utils #
#########

def is_usable_abbr(value):
	if value in DEFAULT_ABBRS:
		raise argparse.ArgumentTypeError(
			f"'{value}' is a default abbreviation. Please choose another."
		)

	return value


def replace_abbreviations(value):
	announce_list = []

	def process_trackers(trackers):
		random_trackers = DEFAULT_TRACKERS.copy()

		tier_list = []
		for item in trackers:
			if isinstance(item, list):
				process_trackers(item)
			elif item == 'open':
				for tracker in DEFAULT_TRACKERS:
					announce_list.append([tracker])
			elif item == 'random':

				tier_list.append(random_trackers.pop())
			else:
				tier_list.append(ABBRS.get(item, item))

		if tier_list:
			announce_list.append(tier_list)

	process_trackers([tier.split('^') for tier in value])

	return announce_list


########
# Meta #
########

meta = create_parser_meta(__title__, __version__)


########
# Show #
########

# Files

show_files = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False
)

show_files_options = show_files.add_argument_group("Show")
show_files_options.add_argument(
	'--show-files',
	action='store_true',
	help="Show files in the summary."
)
show_files_options.add_argument(
	'--hide-files',
	action='store_true',
	help="Don't show files in the summary."
)

# Progress

show_progress = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False,
)

show_progress_options = show_progress.add_argument_group("Show")
show_progress_options.add_argument(
	'--show-progress',
	action='store_true',
	help="Show hashing progress bar."
)
show_progress_options.add_argument(
	'--hide-progress',
	action='store_true',
	help="Hide hashing progress bar."
)


#########
# Local #
#########

local = create_parser_local()


##########
# Filter #
##########

filter_dates = create_parser_filter_dates()


###########
# Torrent #
###########

torrent = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False
)
torrent_options = torrent.add_argument_group("Torrent")
torrent_options.add_argument(
	'--piece-size',
	metavar='SIZE',
	help=(
		"Set piece size.\n"
		"Defaults to automatic calculation.\n"
		f"({', '.join(PIECE_SIZE_STRINGS)})"
	)
)
torrent_options.add_argument(
	'--piece-threshold',
	metavar='TRESHOLD',
	type=int,
	help=(
		"Set piece count threshold for automatic piece size calculation.\n"
		"Defaults to 2000."
	)
)
torrent_options.add_argument(
	'--created-by',
	metavar='CREATOR',
	help=(
		"Set created by field.\n"
		f"Defaults to '{__title__} {__version__}'."
	)
)
torrent_options.add_argument(
	'-c', '--comment',
	metavar='COMMENT',
	help="Set comment field."
)
torrent_options.add_argument(
	'-p', '--private',
	action='store_true',
	help="Make torrent private."
)
torrent_options.add_argument(
	'-P', '--public',
	action='store_true',
	help="Make torrent public."
)
torrent_options.add_argument(
	'-s', '--source',
	metavar='SOURCE',
	help="Set source field."
)
torrent_options.add_argument(
	'--md5',
	action='store_true',
	help="Add md5 hash to info dict."
)


##########
# Output #
##########

output = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False
)

output_options = output.add_argument_group("Output")
output_options.add_argument(
	'-o', '--output',
	metavar='NAME',
	type=lambda p: custom_path(p),
	help=(
		"Set name of torrent file.\n"
		"Defaults to input file or directory name."
	)
)


#########
# Input #
#########

input_ = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False
)

input_options = input_.add_argument_group("Input")
input_options.add_argument(
	'input',
	metavar='PATH',
	type=lambda p: custom_path(p).resolve(),
	help="File or directory."
)

torrent_input = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False
)

torrent_input_options = torrent_input.add_argument_group("Input")
torrent_input_options.add_argument(
	'torrent',
	metavar='TORRENT',
	type=lambda p: custom_path(p).resolve(),
	help="Torrent file."
)


############
# Trackers #
############

trackers = argparse.ArgumentParser(
	argument_default=argparse.SUPPRESS,
	add_help=False
)

trackers_options = trackers.add_argument_group("Trackers")
trackers_options.add_argument(
	'trackers',
	metavar='TRACKERS',
	nargs='*',
	help=(
		"Tracker tiers are separated by a space.\n"
		"Trackers on the same tier are separated with a carat (^)\n\n"
		"Example: 'tracker1^tracker2' tracker3"
	)
)


##########
# thorod #
##########


thorod = argparse.ArgumentParser(
	prog='thorod',
	description="Collection of torrent utilities.",
	usage=argparse.SUPPRESS,
	parents=[meta],
	formatter_class=SubcommandHelpFormatter,
	add_help=False
)

subcommands = thorod.add_subparsers(
	title="Commands",
	dest='_command',
	metavar="<command>"
)


#########
# ABBRS #
#########

abbrs_command = subcommands.add_parser(
	'abbrs',
	description="List/Add/Remove tracker abbreviations.",
	help="List/Add/Remove tracker abbreviations.",
	usage=argparse.SUPPRESS,
	parents=[
		meta
	],
	formatter_class=SubcommandHelpFormatter,
	add_help=False
)
abbrs_command.set_defaults(func=do_abbrs)

abbrs_subcommands = abbrs_command.add_subparsers(
	title="Commands",
	dest='_subcommand',
	metavar="<subcommand>"
)

abbrs_list_command = abbrs_subcommands.add_parser(
	'list',
	description="List tracker abbreviations.",
	help="List tracker abbreviations.",
	usage=argparse.SUPPRESS,
	parents=[
		meta
	],
	formatter_class=UsageHelpFormatter,
	add_help=False
)

abbrs_add_command = abbrs_subcommands.add_parser(
	'add',
	description="Add tracker abbreviations.",
	help="Add tracker abbreviations.",
	usage="thorod abbrs add [ABBREVIATION] [TRACKER]",
	parents=[
		meta
	],
	formatter_class=UsageHelpFormatter,
	add_help=False
)
abbrs_add_command.add_argument(
	'abbreviation',
	metavar='ABBREVIATION',
	help="Abbreviation to use for tracker."
)
abbrs_add_command.add_argument(
	'tracker',
	metavar='TRACKER',
	help="Tracker to abbreviate."
)

abbrs_remove_command = abbrs_subcommands.add_parser(
	'remove',
	description="Remove tracker abbreviations.",
	help="Remove tracker abbreviations.",
	usage="thorod abbrs remove [ABBREVIATIONS]...",
	parents=[
		meta
	],
	formatter_class=UsageHelpFormatter,
	add_help=False
)
abbrs_remove_command.add_argument(
	'abbreviations',
	metavar='ABBREVIATIONS',
	nargs='*',
	help="Abbreviations to remove."
)


##########
# Create #
##########

create_command = subcommands.add_parser(
	'create',
	description="Create a torrent file.",
	help="Create a torrent file.",
	formatter_class=UsageHelpFormatter,
	usage="thorod create [OPTIONS] [PATH] [TRACKERS]...",
	parents=[
		meta,
		show_progress,
		show_files,
		local,
		filter_dates,
		torrent,
		output,
		input_,
		trackers
	],
	add_help=False
)
create_command.set_defaults(func=do_create)


########
# Info #
########

info_command = subcommands.add_parser(
	'info',
	description="Output information about a torrent file.",
	help="Output information about a torrent file.",
	formatter_class=UsageHelpFormatter,
	usage="thorod info [OPTIONS] [TORRENT]",
	parents=[
		meta,
		show_files,
		torrent_input
	],
	add_help=False
)
info_command.set_defaults(func=do_info)


##########
# Magnet #
##########

magnet_command = subcommands.add_parser(
	'magnet',
	description="Generate a magnet link from a torrent file.",
	help="Generate a magnet link from a torrent file.",
	formatter_class=UsageHelpFormatter,
	usage="thorod magnet [OPTIONS] [TORRENT]",
	parents=[
		meta,
		torrent_input
	],
	add_help=False
)
magnet_command.set_defaults(func=do_magnet)


#########
# xseed #
#########

xseed_command = subcommands.add_parser(
	'xseed',
	description="Copy a torrent for cross-seeding.",
	help="Copy a torrent for cross-seeding.",
	formatter_class=UsageHelpFormatter,
	usage="thorod xseed [OPTIONS] [TORRENT]",
	parents=[
		meta,
		torrent,
		output,
		torrent_input,
		trackers
	],
	add_help=False
)
xseed_command.set_defaults(func=do_xseed)


def check_args(args):
	if all(
		option in args
		for option in ['private', 'public']
	):
		raise ValueError("Use one of --private/--public', not both.")

	if all(
		option in args
		for option in ['show_progress', 'hide_progress']
	):
		raise ValueError("Use one of --show-progress/--hide-progress', not both.")

	if all(
		option in args
		for option in ['show_files', 'hide_files']
	):
		raise ValueError("Use one of --show-files/--hide-files', not both.")

	if (
		'torrent' in args
		and not args.torrent.exists()
	):
		raise ValueError(f"'{args.torrent}' does not exist.")

	if (
		'input' in args
		and not args.input.exists()
	):
		raise ValueError(f"'{args.input}' does not exist.")

	if 'trackers' not in args:
		args.trackers = []
	else:
		args.trackers = replace_abbreviations(args.trackers)


def default_args(args):
	defaults = Namespace()

	if 'hide_progress' in args:
		defaults.show_progress = False
		defaults.hide_progress = True
	else:
		defaults.show_progress = True
		defaults.hide_progress = False

	if 'show_files' in args:
		defaults.show_files = True
		defaults.hide_files = False
	else:
		defaults.show_files = False
		defaults.hide_files = True

	defaults.no_recursion = False
	defaults.max_depth = math.inf
	defaults.exclude_paths = []
	defaults.exclude_regexes = []
	defaults.exclude_globs = []

	if 'private' in args:
		defaults.private = True
		defaults.public = False
	else:
		defaults.private = False
		defaults.public = True

	defaults.piece_size = 'auto'
	defaults.piece_threshold = 2000
	defaults.created_by = f"{__title__} {__version__}"
	defaults.comment = None
	defaults.source = None
	defaults.md5 = False

	if 'input' in args:
		defaults.output = Path(args.input.name + '.torrent').resolve()
	elif 'torrent' in args:
		defaults.output = args.torrent.with_name(
			str(args.torrent.name).replace('.torrent', '') + '-xseed.torrent'
		)

	config_defaults = get_defaults(
		args._command,
		read_config_file(),
		command_keys=COMMAND_KEYS
	)

	for k, v in config_defaults.items():
		if k == 'max_depth':
			defaults.max_depth = int(v)
		elif k == 'private':
			defaults['private'] = True
			defaults['public'] = False
		elif k == 'public':
			defaults['public'] = True
			defaults['public'] = False
		elif k in [
			'show_progress',
			'show_files'
		]:
			defaults[k] = v
			defaults[k.replace('show', 'hide')] = not v
		elif k in [
			'hide_progress',
			'hide_files'
		]:
			defaults[k] = v
			defaults[k.replace('hide', 'show')] = not v
		elif k.startswith(('created', 'modified')):
			if k.endswith('in'):
				defaults[k] = datetime_string_to_time_period(v, in_=True)
			elif k.endswith('on'):
				defaults[k] = datetime_string_to_time_period(v, on=True)
			elif k.endswith('before'):
				defaults[k] = datetime_string_to_time_period(v, before=True)
			elif k.endswith('after'):
				defaults[k] = datetime_string_to_time_period(v, after=True)
		else:
			defaults[k] = v

	return defaults


def run():
	try:
		parsed = parse_args(thorod)

		if parsed._command is None:
			thorod.parse_args(['-h'])
		elif parsed._command == 'abbrs':
			parsed.func(parsed)
		else:
			check_args(parsed)
			defaults = default_args(parsed)
			args = merge_defaults(defaults, parsed)

			if args.get('no_recursion'):
				args.max_depth = 0

			args.func(args)
	except KeyboardInterrupt:
		thorod.exit(130, "\nInterrupted by user")



import random

from sortedcontainers import SortedDict


DEFAULT_ABBRS = SortedDict(
	{
		'coppersurfer': 'udp://tracker.coppersurfer.tk:6969/announce',
		'cyberia': 'udp://tracker.cyberia.is:6969/announce',
		'demonii': 'udp://open.demonii.si:1337/announce',
		'desync': 'udp://exodus.desync.com:6969/announce',
		'explodie': 'udp://explodie.org:6969/announce',
		'internetwarriors': 'udp://tracker.internetwarriors.net:1337/announce',
		'itzmx': 'udp://tracker1.itzmx.com:8080/announce',
		'leechers-paradise': 'udp://tracker.leechers-paradise.org:6969/announce',
		'openbittorrent': 'udp://tracker.openbittorrent.com:80/announce',
		'opentrackr': 'udp://tracker.opentrackr.org:1337/announce',
		'port443': 'udp://tracker.port443.xyz:6969/announce',
		'rargb': 'udp://9.rarbg.to:2710/announce',
		'stealth': 'udp://open.stealth.si:80/announce',
		'thetracker': 'udp://thetracker.org:80/announce',
		'torrentclub': 'udp://torrentclub.tech:6969/announce',
		'zer0day': 'udp://tracker.zer0day.to:1337/announce',
	}
)

DEFAULT_TRACKERS = list(DEFAULT_ABBRS.values())
random.shuffle(DEFAULT_TRACKERS)


B = 1024 ** 0
KIB = 1024 ** 1
MIB = 1024 ** 2
GIB = 1024 ** 3
TIB = 1024 ** 4

SYMBOLS = [
	(TIB, 'TiB'),
	(GIB, 'GiB'),
	(MIB, 'MiB'),
	(KIB, 'KiB'),
	(B, 'B'),
]

PIECE_SIZE_VALUES = [
	16 * KIB,
	32 * KIB,
	64 * KIB,
	128 * KIB,
	256 * KIB,
	512 * KIB,
	1 * MIB,
	2 * MIB,
	4 * MIB,
	8 * MIB,
	16 * MIB,
	32 * MIB,
]

PIECE_SIZE_STRINGS = [
	'16k',
	'32k',
	'64k',
	'128k',
	'256k',
	'512k',
	'1m',
	'2m',
	'4m',
	'8m',
	'16m',
	'32m',
	'auto',
]

PIECE_SIZES = dict(zip(PIECE_SIZE_STRINGS, PIECE_SIZE_VALUES))



import os
import random
import string
from hashlib import sha1

from . import bencode
from .constants import PIECE_SIZE_VALUES


def calculate_data_size(files):
	"""Calculate the total size of the input data."""

	return sum(os.path.getsize(f) for f in files)


def calculate_piece_size(data_size, *, threshold=2000):
	for piece_size in PIECE_SIZE_VALUES:
		if data_size / piece_size < threshold:
			break

	return piece_size


def calculate_torrent_size(torrent_info):
	"""Calculate the total size of the files in a torrent."""

	files = torrent_info.get('info').get('files') or [
		torrent_info['info']
	]

	return sum(f['length'] for f in files)


def generate_unique_string():
	"""Generate a random string to make a torrent's infohash unique."""

	return ''.join(
		random.choice(string.ascii_letters + string.digits)
		for x in range(32)
	)


def get_file_path(filepath, basedir):
	"""Get all parts of the file path relative to the base directory of the torrent."""

	return list(filepath.relative_to(basedir).parts)


def hash_info_dict(info_dict):
	return sha1(bencode.dumps(info_dict)).hexdigest()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['thorod']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4,<2.0',
 'attrs>=18.2,<19.4',
 'colorama<1.0.0',
 'pendulum>=2.0,<=3.0,!=2.0.5,!=2.1.0',
 'pprintpp<1.0.0',
 'rich>=0.8',
 'sortedcontainers>=2.0,<3.0',
 'tbm-utils>=2.4,<3.0',
 'tomlkit>=0.5,<0.6']

extras_require = \
{'dev': ['flake8>=3.5,<4.0',
         'flake8-builtins>=1.0,<2.0',
         'flake8-comprehensions>=2.0,<=4.0',
         'flake8-import-order>=0.18,<0.19',
         'flake8-import-order-tbm>=1.0,<2.0',
         'nox>=2019,<2020',
         'sphinx>=2.0,<3.0',
         'sphinx-argparse>=0.2,<0.3',
         'sphinx-material<1.0.0'],
 'doc': ['sphinx>=2.0,<3.0',
         'sphinx-argparse>=0.2,<0.3',
         'sphinx-material<1.0.0'],
 'lint': ['flake8>=3.5,<4.0',
          'flake8-builtins>=1.0,<2.0',
          'flake8-comprehensions>=2.0,<=4.0',
          'flake8-import-order>=0.18,<0.19',
          'flake8-import-order-tbm>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['thorod = thorod.cli:run']}

setup_kwargs = {
    'name': 'thorod',
    'version': '2.1.0',
    'description': 'A CLI utility for torrent creation and manipulation.',
    'long_description': "# thorod\n\n[![PyPI](https://img.shields.io/pypi/v/thorod.svg?label=PyPI)](https://pypi.org/project/thorod/)\n![](https://img.shields.io/badge/Python-3.6%2B-blue.svg)  \n[![GitHub CI](https://img.shields.io/github/workflow/status/thebigmunch/thorod/CI?label=GitHub%20CI)](https://github.com/thebigmunch/thorod/actions?query=workflow%3ACI)  \n[![Docs - Stable](https://img.shields.io/readthedocs/thorod/stable.svg?label=Docs%20%28Stable%29)](https://thorod.readthedocs.io/en/stable/)\n[![Docs - Latest](https://img.shields.io/readthedocs/thorod/latest.svg?label=Docs%20%28Latest%29)](https://thorod.readthedocs.io/en/latest/)\n\n[thorod](https://github.com/thebigmunch/thorod) is a CLI utility for torrent creation and manipulation.\n\n## What's a thorod?\n\nThorod means torrent (of water) in the Tolkien Elvish language of Sindarin.\n\n## Why use thorod?\n\nThere are many CLI torrent utilities out there, so here are some unique or notable features of thorod:\n\n* All torrents are unique; a random salt is added to all created/xseeded torrents.\n* Supports trackers on the same tier.\n* Type less with tracker abbreviations.\n\t* Includes a number of open public trackers by default.\n\t* Includes auto generated open and random abbreviations to help balance load between open public trackers.\n\t* Users can list/add/remove their own tracker abbreviations directly from CLI as well as manually editing config file.\n* Generate magnet links on creation or on command.\n* Has an xseed command to generate a cross-seedable torrent without re-hashing files.\n* View information about a torrent file in the terminal, rather than adding it to a torrent client.\n* Simple automatic piece size calculation from 16 KiB to 32 MiB on by default. Users can set manually by option.\n* Supports source key in info dict used by private trackers.\n\n\n## Installation\n\n``pip install -U thorod``\n\n\n## Usage\n\nFor the release version, see the [stable docs](https://thorod.readthedocs.io/en/stable/).  \nFor the development version, see the [latest docs](https://thorod.readthedocs.io/en/latest/).\n\n\n## Appreciation\n\nShowing appreciation is always welcome.\n\n#### Thank\n\n[![Say Thanks](https://img.shields.io/badge/thank-thebigmunch-blue.svg?style=flat-square)](https://saythanks.io/to/thebigmunch)\n\nGet your own thanks inbox at [SayThanks.io](https://saythanks.io/).\n\n#### Contribute\n\n[Contribute](https://github.com/thebigmunch/thorod/blob/master/.github/CONTRIBUTING.md) by submitting bug reports, feature requests, or code.\n\n#### Help Others/Stay Informed\n\n[Discourse forum](https://forum.thebigmunch.me/)\n\n#### Referrals/Donations\n\n[![Digital Ocean](https://img.shields.io/badge/Digital_Ocean-referral-orange.svg?style=flat-square)](https://bit.ly/DigitalOcean-tbm-referral) [![Namecheap](https://img.shields.io/badge/Namecheap-referral-orange.svg?style=flat-square)](http://bit.ly/Namecheap-tbm-referral) [![PayPal](https://img.shields.io/badge/PayPal-donate-brightgreen.svg?style=flat-square)](https://bit.ly/PayPal-thebigmunch)\n",
    'author': 'thebigmunch',
    'author_email': 'mail@thebigmunch.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/thebigmunch/thorod',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""A simple bencoding implementation with standard API."""

from collections.abc import Mapping
from io import SEEK_CUR, BytesIO


def _bytes(data):
	return bytes(str(data), 'utf8')


def _str(data):
	try:
		return data.decode("utf8")
	except UnicodeDecodeError:
		return data


def _read_until(data, end=b'e'):
	buffer = bytearray()
	d = data.read(1)

	while d != end:
		buffer += d
		d = data.read(1)

	return buffer


def _bdecode_dict(data):
	result = {}

	key = _bdecode(data)

	while key:
		result[key] = _bdecode(data)

		key = _bdecode(data)

	return result


def _bdecode_int(data):
	return int(_read_until(data))


def _bdecode_list(data):
	result = []

	item = _bdecode(data)

	while item:
		result.append(item)

		item = _bdecode(data)

	return result


def _bdecode_str(data):
	length = int(_read_until(data, b':'))

	item = data.read(length)

	return _str(item)


def _bdecode(data):
	type_char = data.read(1)

	if type_char == b'i':
		return _bdecode_int(data)
	elif type_char.isdigit():
		data.seek(-1, SEEK_CUR)
		return _bdecode_str(data)
	elif type_char == b'l':
		return _bdecode_list(data)
	elif type_char == b'd':
		return _bdecode_dict(data)


def _bencode(data):
	if isinstance(data, int):
		return _bytes(f'i{data}e')
	elif isinstance(data, str):
		length = len(_bytes(data))
		return _bytes(f'{length}:{data}')
	elif isinstance(data, (bytes, bytearray)):
		return _bytes(len(data)) + b':' + data
	elif isinstance(data, list):
		return b'l' + b''.join(_bencode(d) for d in data) + b'e'
	elif isinstance(data, Mapping):
		enc_dict = bytes()

		for key in sorted(data):
			enc_dict += _bencode(key) + _bencode(data[key])

		return b'd' + enc_dict + b'e'
	else:
		raise TypeError(
			f"{type(data)} is not a valid type for bencoding."
		)


def dump(obj, fp):
	fp.write(_bencode(obj))


def dumps(obj):
	return _bencode(obj)


def load(fp):
	return _bdecode(fp)


def loads(data):
	if isinstance(data, str):
		data = data.encode()

	return _bdecode(BytesIO(data))



from hashlib import md5, sha1

from sortedcontainers import SortedDict

from . import bencode
from .output import (
	PROGRESS,
	render,
)
from .utils import (
	generate_unique_string,
	get_file_path,
)


def create_dir_info_dict(
	base_path,
	filepaths,
	data_size,
	piece_size,
	private,
	source,
	include_md5,
	*,
	show_progress=True,
):
	def hash_files(progress=None, task=None):
		data = bytes()
		file_infos = []
		pieces = bytearray()

		for filepath in filepaths:
			file_dict = SortedDict()
			length = 0

			md5sum = md5() if include_md5 else None

			with open(filepath, 'rb') as f:
				while True:
					piece = f.read(piece_size)

					if not piece:
						break

					length += len(piece)

					data += piece

					if len(data) >= piece_size:
						pieces += sha1(data[:piece_size]).digest()
						data = data[piece_size:]

					if include_md5:
						md5sum.update(piece)

					if progress:
						progress.update(
							task,
							advance=len(piece),
						)

			file_dict['length'] = length
			file_dict['path'] = get_file_path(filepath, base_path)

			if include_md5:
				file_dict['md5sum'] = md5sum.hexdigest()

			file_infos.append(file_dict)

		if len(data) > 0:
			pieces += sha1(data).digest()

		return file_infos, pieces

	if show_progress:
		render("\n Hashing Files\n\n", style="bold yellow")

		with PROGRESS:
			task = PROGRESS.add_task(
				"Hashing",
				total=data_size,
			)
			file_infos, pieces = hash_files(PROGRESS, task)
	else:
		file_infos, pieces = hash_files()

	info_dict = SortedDict()
	info_dict['files'] = file_infos
	info_dict['name'] = base_path.name
	info_dict['pieces'] = pieces
	info_dict['piece length'] = piece_size
	info_dict['salt'] = generate_unique_string()

	info_dict['private'] = 1 if private else 0

	if source:
		info_dict['source'] = source

	return info_dict


def create_file_info_dict(
	filepaths,
	data_size,
	piece_size,
	private,
	source,
	include_md5,
	show_progress=True,
):
	def hash_file(progress=None, task=None):
		pieces = bytearray()
		length = 0
		md5sum = md5() if include_md5 else None

		with open(filepaths[0], 'rb') as f:
			while True:
				piece = f.read(piece_size)

				if not piece:
					break

				length += len(piece)

				pieces += sha1(piece).digest()

				if include_md5:
					md5sum.update(piece)

				if progress:
					progress.update(
						task,
						advance=len(piece),
					)

		return pieces, length, md5sum

	if show_progress:
		render("\n Hashing Files\n\n", style="bold yellow")

		with PROGRESS:
			task = PROGRESS.add_task(
				"Hashing",
				total=data_size,
			)

			pieces, length, md5sum = hash_file(PROGRESS, task)
	else:
		pieces, length, md5sum = hash_file()

	info_dict = SortedDict()
	info_dict['name'] = filepaths[0].name
	info_dict['length'] = length
	info_dict['pieces'] = pieces
	info_dict['piece length'] = piece_size
	info_dict['salt'] = generate_unique_string()

	info_dict['private'] = 1 if private else 0

	if source:
		info_dict['source'] = source

	if include_md5:
		info_dict['md5sum'] = md5sum.hexdigest()

	return info_dict


def read_torrent_file(filepath):
	try:
		torrent_info = bencode.load(filepath.open('rb'))
	except FileNotFoundError:
		raise FileNotFoundError(f"{filepath} not found.")
	except TypeError:
		raise TypeError(f"Could not parse {filepath}.")

	return torrent_info


def write_torrent_file(filepath, torrent_info):
	bencode.dump(torrent_info, filepath.open('wb'))
